
/*
 * $Id: Client.cc,v 1.35 1999/09/07 17:46:00 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "CltConnMgr.h"
#include "CltOpts.h"
#include "CltXact.h"
#include "Client.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "ErrorMgr.h"

#include "RobotSym.h"

#include "polyBcastChannels.h"
#include "polyErrors.h"
#include "globals.h"


ObjFarm<CltXact> Client::TheXacts;

Client::Client(): theOriginDistr(0), theLaunchCount(0) {
}

Client::~Client() {
}

void Client::configure(const RobotSym *cfg) {
	SockOpt opt;
	Agent::configure(cfg, opt);
	theHost.port(-1); // we do not care about port number

	configureConnMgrs(cfg, opt);

	theLaunchWin = cfg->launchWin();

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgrs.count());

	Agent::start();

	if (theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);
	ConnMgr *mgr = x->conn()->mgr();
	Assert(mgr);
	mgr->put(x->conn());
	TheXacts.put(x);
}

void Client::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	if (theThinkDistr) {
		const double lf = TheStatPhaseMgr->loadFactor(lastLaunch);
		Assert(lf > 0);
		double delay;
		if (TheStatPhaseMgr->varyLoad()) {
			// next launch should happen [0, 2dt] from last launch
			// where dt is 1/current_rate
			static RndGen rng;
			delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
		} else {
			delay = theThinkDistr->trial()/lf;
		}
		sleepTill(lastLaunch + Time::Secd(delay));
	} else
		launch();
}

void Client::launch() {
	// select origin server to connect to
	const int sid = (int)theOriginDistr->trial();
	Assert(sid >= 0 && sid < theConnMgrs.count());

	// connect and execute transaction
	if (Connection *conn = theConnMgrs[sid]->get()) {
		theLaunchCount++;
		TheXacts.get()->exec(this, conn);
	} else {
		if (ExplainError(errno)) {
			Comment(1) << here << theHost << " failed to connect to " 
				<< theConnMgrs[sid]->addr() << "; " << error(errno)
				<< endc;
		}
		Broadcast(TheXactErrChannel, (Xaction*)0);

		// if no alarms are pending, we may get stuck and submit
		// no more xactions (e.g. best effort); this is a kludge
		if (!thePendAlarmCnt)
			sleepTill(TheClock); // submit more next cycle
	}
}

void Client::configureConnMgrs(const RobotSym *cfg, const SockOpt &opt) {
	Array<NetAddr> addrs;
	if (!cfg->origins(addrs, theOriginDistr) || !addrs.count()) {
		Comment(1) << ThePrgName << ": no origin addresses specified for robot " << cfg->kind() << endc;
		FatalError(errOther);
	}
	Assert(theOriginDistr);

	// create and configure each manager
	RndDistr *d = cfg->pconnUseLmt();
	for (int i = 0; i < addrs.count(); ++i) {
		CltConnMgr *mgr = new CltConnMgr;
		const NetAddr &connAddr = TheCltOpts.thePxyAddr ? 
			TheCltOpts.thePxyAddr : addrs[i];
		mgr->configure(connAddr, addrs[i], opt, -1, d);
		theConnMgrs.append(mgr);
	}
}
