
/*
 * $Id: CltXact.cc,v 1.18 1999/09/07 17:46:02 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <strstream.h>

#include "Clock.h"
#include "Connection.h"
#include "StatPhaseMgr.h"
#include "SharedOpts.h"
#include "World.h"
#include "ConnMgr.h"
#include "ErrorMgr.h"
#include "CltOpts.h"
#include "Client.h"
#include "CltXact.h"

#include "polyErrors.h"
#include "gadgets.h"
#include "globals.h"


/* HTTP header strings (to avoid strlen() calls when creating headers) */

// request line parts
static const String rlpGetHttpOrg = "GET /http:/";       
static const String rlpGetHttpPxy = "GET http://";
static const String rlpHttp_10    = " HTTP/1.0\r\n";

// full header fields
static const String hfAccept       = "Accept: */*\r\n";
static const String hfConnAliveOrg = "Connection: keep-alive\r\n";
static const String hfConnCloseOrg = "Connection: close\r\n";
static const String hfConnAlivePxy = "Proxy-Connection: keep-alive\r\n";
static const String hfConnClosePxy = "Proxy-Connection: close\r\n";

// header field prefixes
static const String hfpHost  = "Host: ";
static const String hfpXXact = "X-Xact: ";



void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theConsumedSize = 0;
	theState = stNone;

	theRepHdr.reset();
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);

	theId.create();
	makeReq(theConn->theWrBuf);
	theState = stSpaceWaiting;
	theConn->theWr.start(this);
}

void CltXact::finish(int errNo) {
	stop(errNo);
	theOwner->noteXactDone(this);
}

void CltXact::noteReadReady(int) {
	theConn->read();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theState == stHdrWaiting) {
		if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {

			// dump reply headers if needed
			if (TheOpts.doPrintReps)
				cout.write(theConn->theRdBuf.content(), theRepHdr.theHdrSize);

			if (const int errNo = interpretHeader()) {
				finish(errNo);
				return;
			}
		} else
		if (theConn->theRdBuf.full()) { // too big header
			finish(errHugeHdr);
			return;
		} else
		if (theConn->atEof()) { // premature end of headers
			finish(errPrematureEoh);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		// consume up to server's debt()
		const Size debt = theRepSize - theConsumedSize;
		if (debt < theConn->theRdBuf.contSize()) {
			theConsumedSize = theRepSize; // += debt;
			theConn->theRdBuf.consumed(debt);
		} else {
			theConsumedSize += theConn->theRdBuf.contSize();
			theConn->theRdBuf.consumedAll();
		}

		if (theConsumedSize == theRepSize) { // got all we need
			// account for foreign replies here
			if (!theRepHdr.theXactId)
				finish(errForeignRep);
			else
			if (theConn->theRdBuf.contSize() && !theConn->reusable())
				finish(errExtraRepData);
			else
				finish(0);
			return;
		}
	}

	if (theConn->exhausted()) { // premature eof
		finish(errPrematureEof);
		return;
	}
}

void CltXact::noteWriteReady(int) {

	Assert(theConn->theWrBuf.contSize() > 0);
	theConn->write();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theConn->theWrBuf.empty()) {
		theConn->theWr.stop(this);
		theState = stHdrWaiting;
		theConn->theRd.start(this);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

int CltXact::interpretHeader() {

	if (theRepHdr.theContSize < 0) // no content len
		return errNoCLen;

	theRepSize = theRepHdr.theHdrSize + theRepHdr.theContSize;

	// note: read foreign replies, but
	// do not count foreign replies as misses, hits, etc.!
	if (theRepHdr.theXactId) {

		theOidRec.cachable = theRepHdr.isCachable;
		theOidRec.hit = !theId.myMutant(theRepHdr.theXactId);

		if (!theOidRec.repeat && theOidRec.hit && !TheCltOpts.ignoreFalseHits)
			return errFalseHit;

		if (!theOidRec.cachable && theOidRec.hit)
			return errUnchbHit;

		// hmm, foreigners are not noted?
		TheWorld.noteResponse(theOidRec);

		if (TheCltOpts.printFalseMisses && theOidRec.repeat && theOidRec.cachable && !theOidRec.hit) {
			cout << "False Miss Oid: " << theOidRec.oid << endl;
			cout.write(theConn->theRdBuf.content(), theRepHdr.theHdrSize);
		}
	}

	// note if the server will close the connection
	theConn->lastUse(!theRepHdr.keepAlive());

	theState = stBodyWaiting;
	return 0;
}

void CltXact::makeReq(WrBuf &buf) {
	static int oid_off = Max(0, TheCltOpts.theWorld.theUniqStart - 1);
	const int oid = TheWorld.oid(oid_off, theOidRec);

	const char *reqStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	// request line prefix

	// make absolute and relative urls of the same length
	// so transparent folks do not get an advantage
	os << (TheCltOpts.thePxyAddr ? rlpGetHttpPxy : rlpGetHttpOrg);

	const NetAddr &orig = theConn->mgr()->originAddr();
	os << orig.addrA();
	if (orig.port() != 80)
		os << ':' << orig.port();

	// url
	if (const char *url = TheWorld.url(oid)) {
		os << url; // preloaded
	} else {
		os 
			<< '/' << '_' << oid 
			<< '/' << TheWorld.id() 
			<< '/' << TheWorld.type()
			;

		if (TheCltOpts.theWorld.theUniqStart > 0)
			os << '/' << TheWorld.uid(oid_off);
	}

	os << rlpHttp_10; // end-of-request-line

	/*
	 * header fields: it is "good practice" to send general-header
	 * fields first, followed by request-header or response- header
	 * fields, and ending with the entity-header fields.
	 */

	// general-header fields
	os << (TheCltOpts.thePxyAddr ?
		(theConn->reusable() ? hfConnAlivePxy : hfConnClosePxy) :
		(theConn->reusable() ? hfConnAliveOrg : hfConnCloseOrg));

	// request-header fields
	os
		<< hfAccept
		<< hfpHost << orig.addrA() << ':' << orig.port() << crlf
		;

	// entity-header field
	os << hfpXXact << theId << crlf;

	os << crlf; // end-of-header

	oid_off++;
	buf.appended(os.tellp());

	// dump request
	if (TheOpts.doPrintReqs)
		cout.write(reqStart, buf.space()-reqStart);
}
