
/*
 * $Id: World.cc,v 1.7 1999/09/07 17:46:34 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <math.h>
#include <fstream.h>

#include "PopModel.h"
#include "TmlModel.h"
#include "HitModel.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "World.h"
#include "gadgets.h"

World TheWorld;

World::World(): thePopMdl(0), theTmlMdl(0), theHitMdl(0), theLastId(0) {
	// thePopMdl must be set by the client (there may be several models)
	theTmlMdl = new TmlModel;
	theHitMdl = new HitModel;
}

World::~World() {
	while (theUrls.count())
		delete[] theUrls.pop();
	delete thePopMdl;
	delete theTmlMdl;
	delete theHitMdl;
}


int World::genOid(OidRec &rec) {
	Assert(thePopMdl);

	replanOid = true;

	// repeat?
	rec.repeat = theLastId && theHitMdl->hitIt();

	const int oid = rec.repeat ? repeatOid(theLastId) : ++theLastId;

	// update history
	if (theTmlMdl->enabled()) {
		if (replanOid)
			theTmlMdl->addOid(oid);
		theTmlMdl->advance();
	}

	Assert(oid > 0);
	rec.oid = oid;
	return oid;
}

// generate an old Oid using tmp locality model (if any)
int World::repeatOid(int lastId) {
	if (theTmlMdl->enabled()) {
		if (const int oid = theTmlMdl->pickOid())
			return oid;
	}

	// XXX: this "replan" logic is very confusing
	replanOid = false;
	return thePopMdl->repeatId(lastId);
}

void World::prefill(int count) {
	Assert(count > 0);
	Assert(!theLastId); // we have not started generating Ids yet
	theLastId = count;

	// let the popularity and tmp locality models to prefill the plan
	if (theTmlMdl->enabled()) {

		// when to stop?!
		// a mean plus three std_dev
		int depth = (int) (theTmlMdl->distr()->mean() + 3*theTmlMdl->distr()->sdev());
		// adjust for hit rate
		depth += (int) (depth * (1 - theHitMdl->hr())); 
		
		Comment(3) << "warming up the tmp loc plan; oids: " << count << " slots: " << depth << endc;

		while (depth-- > 0) {
			// XXX: this emulates genOid except we do not add misses
			if (theHitMdl->hitIt())
				theTmlMdl->addOid(repeatOid(theLastId));
			theTmlMdl->advance();
		}

		Comment(3) << "warmed up the tmp loc plan; current depth: " << theTmlMdl->curPlanDepth() << endc;
	}
}

bool World::loadUrls(const char *fname) {
	char url[8*1024];

	Assert(fname);

	ifstream is(fname);
	Size sz = 0;
	int lineCount = 0;
	int urlCount = 0;

	Comment(1) << "loading urls from '" << fname << "'" << endc;

	while (is && is.getline(url, sizeof(url))) {
		lineCount++;
		const int len = strlen(url);
		if (len && *url == '/') {
			theUrls.append(StrDup(url));
			sz += sizeof(char*) + len + 1;
			urlCount++;
		} else
		if (len && *url != '#') {
			Comment << fname << ':' << lineCount << " unknown url format near " << url << endc;
			return false;
		}
	}

	is.close();

	Comment(1)
		<< "loaded " << urlCount 
		<< " urls from '" << fname << "'"
		<< " (" << lineCount << " lines)"
		<< " consumed ~ " << ((int)sz)/(1024*1024.) << " MB" << endc;

	return urlCount > 0;
}

ostream &World::report(ostream &os) const {
	return os << "Unique oids: " << theLastId+1;
}
