
/*
 * $Id: httpHdrs.cc,v 1.7 1999/09/07 17:46:38 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "String.h"
#include "StrIdentifier.h"
#include "HttpDate.h"
#include "httpHdrs.h"
#include "gadgets.h"


/* internal type to store static parsing info */
class MsgHdrParsTab {
	public:
		MsgHdrParsTab();
		~MsgHdrParsTab();

	public:
		StrIdentifier *ids;
		Array<MsgHdr::Parser> *parsers;
};


MsgHdrParsTab *ReqHdr::TheParsTab = 0;
MsgHdrParsTab *RepHdr::TheParsTab = 0;


// strNchr(), sort of
static
const char *strBoundChr(const char *s, char c, const char *eos) {
	while (s < eos) {
		if (*s == c)
			return s;
		else
			++s;
	}
	return 0;
}


/* MsgHdr */

MsgHdr::MsgHdr(const MsgHdrParsTab &aTab): theParsTab(aTab) {
	reset();
}

void MsgHdr::reset() {
	theHdrSize = 0;
	theContSize = -1;
	theXactId.clear();
	theHttpVer = -1;
	theKeepAliveConn = kaDefault;

	theBufBeg = theBufEnd = theSrchPtr = 0;
	theFields.reset();
	theSrchState = ssFirst;

	// do not reset parsing tables
}

bool MsgHdr::keepAlive() const {
	if (theHttpVer >= 1.1) // 1.1: keep unless told otherwise
		return theKeepAliveConn != kaNo;
	else                   // 1.0: keep if explicitly told so
		return theKeepAliveConn == kaYes;
}

// note: buf does not have to be zero-terminated!
bool MsgHdr::parse(const char *buf, Size sz) {
	if (!theBufBeg) { // have not started the search yet
		theBufBeg = theSrchPtr = buf;
		Assert(theSrchState == ssFirst);
	} else {           // continue search
		Assert(theBufBeg == buf);
		Assert(theSrchState != ssFound);
	}

	theBufEnd = buf + sz; // to be refined later
	do {
		// search for LF
		if (theSrchState == ssFirst) {
			do {
				if (*theSrchPtr++ == '\n') {
					theSrchState = ssSkip;
					break;
				}
			} while (theSrchPtr < theBufEnd);
		}

		// LF after skipping optional CRs means and-of-headers
		while (theSrchState == ssSkip && theSrchPtr < theBufEnd) {
			if (*theSrchPtr == '\n') {
				theSrchState = ssFound;
			} else
			if (*theSrchPtr != '\r') {
				theFields.append(theSrchPtr); // start of a header!
				theSrchState = ssFirst;
			}
			++theSrchPtr;
		}
	} while (theSrchPtr < theBufEnd && theSrchState != ssFound);

	if (theSrchState != ssFound)
		return false;

	// found end-of-headers!
	theBufEnd = theSrchPtr;
	theHdrSize = theBufEnd - theBufBeg;

	// now parse known fields
	// luckily, we already know field starts!
	parseFields();
	return true;
}

void MsgHdr::parseFields() {
	const char *eoh = theBufEnd;
	parseRLine(theBufBeg, theFields.count() ? theFields[0] : theBufEnd);

	for (int i = theFields.count()-1; i >= 0; --i) {
		const char *hdr = theFields[i];
		const int len = eoh-hdr; // approximate (includes crlfs)
		const int id = theParsTab.ids->lookup(hdr, len);
		if (id > 0) {
			const char *val = hdr + theParsTab.ids->string(id).len();
			while (isspace(*val)) ++val;
			Parser p = theParsTab.parsers->item(id);
			(this->*p)(val, eoh);
		}
		eoh = hdr;
	}
}

bool MsgHdr::parseContLen(const char *buf, const char *) {
	theContSize = xatoi(buf, -1);
	return theContSize >= 0;
}

bool MsgHdr::parseXXact(const char *buf, const char *eoh) {
	if (const char *b = strBoundChr(buf, '.', eoh))
		if (const char *c = strBoundChr(++b, ':', eoh))
				return theXactId = UniqId(xatoi(buf), xatoi(b), xatoi(c+1));
	return false;
}


/* XXX: Connection and other headers may have a _list_ of options */

bool MsgHdr::parseConnection(const char *buf, const char *) {
	if (!strncasecmp("close", buf, 5))
		theKeepAliveConn = kaNo;
	else
	if (!strncasecmp("keep", buf, 4))
		theKeepAliveConn = kaYes;
	else
		return false;
	return true;
}

// adds definitions common to replies and requests
void MsgHdr::Configure(MsgHdrParsTab &tab) {
	AddParser("Content-Length:", &MsgHdr::parseContLen, tab);
	AddParser("X-Xact:", &MsgHdr::parseXXact, tab);
	AddParser("Connection:", &MsgHdr::parseConnection, tab);
	AddParser("Proxy-Connection:", &MsgHdr::parseConnection, tab);
}

int MsgHdr::AddParser(const String &field, Parser parser, MsgHdrParsTab &where) {
	const int id = where.ids->add(field);
	where.parsers->put(parser, id);
	return id;
}

// these should never be called
inline bool dontCallMe() { Assert(0); return false; }
bool MsgHdr::parsePragma(const char *, const char *)  { return dontCallMe(); }
bool MsgHdr::parseCControl(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseIMS(const char *, const char *) { return dontCallMe(); }


/* ReqHdr */

ReqHdr::ReqHdr(): MsgHdr(*TheParsTab), theObjId(-1) {
}

void ReqHdr::reset() {
	MsgHdr::reset();
	theIMS = Time();
	theObjId = -1;
}

bool ReqHdr::parseRLine(const char *buf, const char *eorl) {
	if (!strncasecmp("GET ", buf, 4)) {
		buf += 4;
		if (const char *h = strBoundChr(buf, '_', eorl))
			theObjId = xatoi(buf = h + 1, -1);
		if (const char *h = strBoundChr(buf, ' ', eorl))
			isDbl(h + 6, theHttpVer);
		return true;
	}
	return false;
}

bool ReqHdr::parseIMS(const char *buf, const char *eoh) {
	theIMS = HttpDateParse(buf, eoh - buf);
	return theIMS >= 0;
}

void ReqHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("If-Modified-Since:", &MsgHdr::parseIMS, *TheParsTab);
}

void ReqHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* RepHdr */

RepHdr::RepHdr(): MsgHdr(*TheParsTab), isCachable(true) {
}

void RepHdr::reset() {
	MsgHdr::reset();
	isCachable = true;
}

bool RepHdr::parseRLine(const char *buf, const char *) {
	// what is the actual version format in the specs??
	if (!strncasecmp("HTTP/", buf, 5))
		return isDbl(buf+5, theHttpVer);
	else
		return false;
}

bool RepHdr::parsePragma(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

bool RepHdr::parseCControl(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

void RepHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("Pragma:", &MsgHdr::parsePragma, *TheParsTab);
	AddParser("Cache-Control:", &MsgHdr::parseCControl, *TheParsTab);
}

void RepHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* MsgHdrParsTab */

MsgHdrParsTab::MsgHdrParsTab() {
	ids = new StrIdentifier;
	parsers = new Array<MsgHdr::Parser>;
}

MsgHdrParsTab::~MsgHdrParsTab() {
	delete ids; ids = 0;
	delete parsers; parsers = 0;
}



/* initialization */

int HttpHdrsInit::TheUseCount = 0;

void HttpHdrsInit::init() {
	ReqHdr::Configure();
	RepHdr::Configure();
}

void HttpHdrsInit::clean() {
	RepHdr::Clean();
	ReqHdr::Clean();
}
