
/*
 * $Id: PglArraySym.cc,v 1.8 1999/09/07 17:47:40 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "pgl.h"

#include "TblDistr.h"
#include "PglArraySym.h"
#include "gadgets.h"


ArraySym::ArraySym(const String &anItemType): 
	ListSym(anItemType ? (const char *)(anItemType + "[]") : "Array"),
	theItemType(anItemType), warnedBadProbs(false) {
}

bool ArraySym::isA(const String &type) const {
	return ListSym::isA(type) || type == "Array";
}

SynSym *ArraySym::clone(const String &type) const {
	String h = type;
	String itemType;
	if (const char *p = h.str("[]"))
		itemType = h(0, p - h);
	else
	if (isA(type))
		itemType = theItemType;
	else
		return 0;

	ArraySym *clone = new ArraySym(itemType);

	if (!clone->absorb(*this)) {
		delete clone;
		return 0;
	}

	clone->syncProbs();
	for (int i = 0; i < clone->count(); ++i)
		clone->prob(i, theProbs[i]);

	return clone;
}

void ArraySym::addProb(const SynSym &s, double p) {
	ListSym::add(s);
	prob(count()-1, p);
}

void ArraySym::addFailure(const SynSym &s) {
	cerr << s.loc() << "cannot add `" << s.type() << "' item to `"
		<< (theItemType ? (const char*) type() : "yet-type-less")
		<< "' array" << endl << xexit;
}

bool ArraySym::cadd(const SynSym &s) {
	// absorb other arrays (mimic Perl list behavior)
	if (s.isA("Array"))
		return absorb((ArraySym&)s.cast("Array"));

	// note: anonymous array has a type of its first item
	const String it = theItemType ? theItemType : s.type();

	if (SynSym *sclone = s.clone(it)) {

		if (!theItemType) {
			theItemType = it;
			type(it + "[]");
		}

		doAdd(sclone, s);
		syncProbs();
		return true;
	}

	return false;
}

ostream &ArraySym::print(ostream &os, const String &pfx) const {
	Array<double> probs;
	makeProbs(probs);
	os << '[';
	for (int i = 0; i < count(); ++i) {
		if (i) os << ',';
		Assert(item(i));
		item(i)->print(os, pfx);
		Assert(probs[i] >= 0);
		os << ':' << (probs[i]*100) << '%';
	}
	return os << ']';
}

void ArraySym::prob(int idx, double p) {
	syncProbs();
	Assert(idx < theProbs.count());
	theProbs[idx] = p;
}

// count the number of undefiend probs and
// get the sum of defined probs
double ArraySym::statProbs(int &undefCnt) const {
	Assert(theProbs.count() == count()); // synced

	undefCnt = 0;
	double sum = 0;
	for (int i = 0; i < theProbs.count(); ++i) {
		if (theProbs[i] > 0)
			sum += theProbs[i];
		else
		if (theProbs[i] < 0)
			undefCnt++;
	}
	return sum;
}

void ArraySym::makeProbs(Array<double> &res) const {
	int undefCnt = 0;
	const double sum = statProbs(undefCnt);

	if (sum > 1.0) {
		cerr << loc() << "array probabilities (below) add up to " << 100*sum << '%' << endl;
		print(cerr << loc()) << endl << xexit;
	}

	// copy values
	res.stretch(theProbs.count());
	res.count(theProbs.count());
	for (int i = 0; i < theProbs.count(); ++i)
		res[i] = theProbs[i];

	// assign sum leftovers to the undefined probs
	if (undefCnt) {
		const double restp = Max((1.0-sum)/undefCnt, 0.0);
		for (int i = 0; i < res.count(); ++i) {
			if (res[i] < 0)
				res[i] = restp;
		}
	}
}

void ArraySym::makeProbs() {
	int undefCnt = 0;
	const double sum = statProbs(undefCnt);
	if (sum < 1.0 && !undefCnt && !warnedBadProbs) {
		print(cerr << loc()) << endl;
		cerr << loc() << "array probabilities (above) add up to " << 100*sum << "% (not 100%) and no wild-card entries are present;"
			<< " missing percents will be spread out proportionally among array entries" << endl;
		warnedBadProbs = true;
	}

	makeProbs(theProbs);
}

// ensure that all probabilities are off by default
void ArraySym::syncProbs() {
	while (theProbs.count() < count())
		theProbs.append(-1);
}

// uses a table distribution to model item probabilities
// selector is a distribution of item positions
RndDistr *ArraySym::makeSelector(const String &name) {
	makeProbs();
	TblDistr *sel = new TblDistr(name);

	for (int i = 0; i < count(); ++i) {
		TblDistrBin *bin = new TblDistrPoint(i);
		bin->absContrib(prob(i));
		sel->addBin(bin);
	}

	sel->make();
	return sel;
}
